@@ -52,7 +52,9 @@ module Data.Conduit
     , mapOutput
     , mapOutputMaybe
     , mapInput
+    , mergeSource
     , passthroughSink
+    , sourceToList
 
       -- * Connect-and-resume
     , ResumableSource
