@@ -1,4 +1,5 @@
 {-# OPTIONS_HADDOCK not-home #-}
+{-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE CPP #-}
@@ -6,8 +7,7 @@
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TupleSections #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE ImpredicativeTypes #-}
+{-# LANGUAGE Trustworthy #-}
 module Data.Conduit.Internal.Conduit
     ( -- ** Types
       ConduitM (..)
@@ -74,7 +74,9 @@ module Data.Conduit.Internal.Conduit
     , zipSources
     , zipSourcesApp
     , zipConduitApp
+    , mergeSource
     , passthroughSink
+    , sourceToList
     , fuseBoth
     , fuseBothMaybe
     , fuseUpstream
@@ -123,13 +125,13 @@ instance Functor (ConduitM i o m) where
     fmap f (ConduitM c) = ConduitM $ \rest -> c (rest . f)
 
 instance Applicative (ConduitM i o m) where
-    pure = return
+    pure x = ConduitM ($ x)
     {-# INLINE pure #-}
     (<*>) = ap
     {-# INLINE (<*>) #-}
 
 instance Monad (ConduitM i o m) where
-    return x = ConduitM ($ x)
+    return = pure
     ConduitM f >>= g = ConduitM $ \h -> f $ \a -> unConduitM (g a) h
 
 instance MonadThrow m => MonadThrow (ConduitM i o m) where
@@ -664,6 +666,25 @@ unwrapResumableConduit (ResumableConduit src final) = do
 newResumableConduit :: Monad m => Conduit i m o -> ResumableConduit i m o
 newResumableConduit (ConduitM c) = ResumableConduit (c Done) (return ())
 
+
+-- | Merge a @Source@ into a @Conduit@.
+-- The new conduit will stop processing once either source or upstream have been exhausted.
+mergeSource
+  :: Monad m
+  => Source m i
+  -> Conduit a m (i, a)
+mergeSource = loop . newResumableSource
+  where
+    loop :: Monad m => ResumableSource m i -> Conduit a m (i, a)
+    loop src0 = await >>= maybe (lift $ closeResumableSource src0) go
+      where
+        go a = do
+          (src1, mi) <- lift $ src0 $$++ await
+          case mi of
+            Nothing -> lift $ closeResumableSource src1
+            Just i  -> yield (i, a) >> loop src1
+
+
 -- | Turn a @Sink@ into a @Conduit@ in the following way:
 --
 -- * All input passed to the @Sink@ is yielded downstream.
@@ -700,6 +721,27 @@ passthroughSink (ConduitM sink0) final = ConduitM $ \rest -> let
                 go [] (next x)
     in go [] (sink0 Done)
 
+-- | Convert a @Source@ into a list. The basic functionality can be explained as:
+--
+-- > sourceToList src = src $$ Data.Conduit.List.consume
+--
+-- However, @sourceToList@ is able to produce its results lazily, which cannot
+-- be done when running a conduit pipeline in general. Unlike the
+-- @Data.Conduit.Lazy@ module (in conduit-extra), this function performs no
+-- unsafe I\/O operations, and therefore can only be as lazily as the
+-- underlying monad.
+--
+-- Since 1.2.6
+sourceToList :: Monad m => Source m a -> m [a]
+sourceToList =
+    go . flip unConduitM Done
+  where
+    go (Done _) = return []
+    go (HaveOutput src _ x) = liftM (x:) (go src)
+    go (PipeM msrc) = msrc >>= go
+    go (NeedInput _ c) = go (c ())
+    go (Leftover p _) = go p
+
 -- Define fixity of all our operators
 infixr 0 $$
 infixl 1 $=
